/* args.c - Parse command-line argument for xsvg using argp
 *
 * Copyright © 2002 USC/Information Sciences Institute
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Information Sciences Institute not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission.  Information Sciences Institute
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * INFORMATION SCIENCES INSTITUTE DISCLAIMS ALL WARRANTIES WITH REGARD
 * TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL INFORMATION SCIENCES
 * INSTITUTE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Carl Worth <cworth@isi.edu>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <getopt.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "args.h"

static const char ARGS_PROGRAM_VERSION[] = VERSION;
static const char ARGS_PROGRAM_DESCRIPTION[] = "X11 SVG viewer";
static const char ARGS_PROGRAM_BUG_ADDRESS[] = "<cworth@cworth.org>";

static const char ARGS_PROGRAM_ARGDOC[] = "<SVG file> (\"-\" for stdin)";

/* XXX: getopt is rather annoying in that you must maintain a
 * string-encoding of the options in addition to the table. For
 * example, to enable the sample display option below, one would have
 * to add "d:" to the optstring.
 *
 * A useful exercise would be to write a function to generate the
 * optstring from the table.
 *
 * The other annoying thing is that the table does not include help
 * strings, so the args_help function below must also be maintained
 * manually.
 */

enum {
    ARGS_VAL_FLIPX = 256,
    ARGS_VAL_FLIPY,
    ARGS_VAL_HELP
};

static char args_optstring[] = "d:g:s:w:h:fFav";
static struct option args_options[] = {
    /* name,		has_arg,	flag,	val */
    {"display",		1,		0,	'd'},
    {"geometry",	1,		0,	'g'},
    {"scale",		1,		0,	's'},
    {"flipx",		0,		0,	ARGS_VAL_FLIPX},
    {"flipy",		0,		0,	ARGS_VAL_FLIPY},
    {"fit",             0,		0,	'f'},
    {"argb",		0,		0,	'a'},
    {"help",		0,		0,	ARGS_VAL_HELP},
    {"version",		0,		0,	'V'},
    { 0 }
};

static void
args_help (const char *argv0)
{
    char *argv0_copy = strdup (argv0);
    char *argv0_base = basename (argv0_copy);

    printf ("Usage: %s [OPTIONS] %s\n", argv0_base, ARGS_PROGRAM_ARGDOC);
    printf ("%s - %s\n", argv0_base, ARGS_PROGRAM_DESCRIPTION);
    puts ("");
    printf ("  -d, --display=DISPLAY  \tX server to contact\n");
    printf ("  -g, --geometry=GEOMETRY\tSize and location of window\n");
    printf ("  -s, --scale=FACTOR     \tScale image by FACTOR\n");
    puts ("");
    printf ("      --flipx            \tFlip X coordinates\n");
    printf ("      --flipy            \tFlip Y coordinates\n");
    printf ("  -f, --fit              \tZoom to fit in window\n");
    printf ("  -a, --argb             \tUse an ARGB window\n");
    puts ("");
    printf ("      --help             \tGive this help list\n");
    printf ("  -V, --version          \tPrint program version\n");

    free (argv0_copy);
};

static void
args_usage (const char *argv0)
{
    char *argv0_copy = strdup (argv0);
    char *argv0_base = basename (argv0_copy);

    printf ("Usage: %s [OPTION] %s\n", argv0_base, ARGS_PROGRAM_ARGDOC);
    printf ("Try `%s --help' for more information.\n", argv0_base);

    free (argv0_copy);
}

int
args_parse (args_t *args, int argc, char *argv[])
{
    int c;

    args->display = NULL;
    args->geometry = NULL;
    args->scale = 1.0;
    args->svg_files = malloc (argc * sizeof (char *));
    args->svg_nfile = 0;

    args->flipx = 0;
    args->flipy = 0;
    args->argb = 0;

    args->fit = 0;

    while (1) {
	c = getopt_long (argc, argv, args_optstring, args_options, NULL);
	if (c == -1)
	    break;

	switch (c) {
	case 'd':
	    args->display = optarg;
	    break;
	case 'g':
	    args->geometry = optarg;
	    break;
	case 's':
	    args->scale = atof (optarg);
	    break;
	case ARGS_VAL_FLIPX:
	    args->flipx = 1;
	    break;
	case ARGS_VAL_FLIPY:
	    args->flipy = 1;
	    break;
	case 'f':
	    args->fit = 1;
	    break;
	case 'a':
	    args->argb = 1;
	    break;
	case 'V':
	    printf ("%s\n", ARGS_PROGRAM_VERSION);
	    exit (0);
	    break;
	case ARGS_VAL_HELP:
	    args_help (argv[0]);
	    exit (0);
	    break;
	case '?':
	    args_help (argv[0]);
	    exit (1);
	    break;
	default:
	    fprintf (stderr, "Unhandled option: %d\n", c);
	    exit (1);
	    break;
	}
    }

    if (optind == argc) {
	args_usage (argv[0]);
	exit (1);
    }

    while (optind < argc)
	args->svg_files[args->svg_nfile++] = argv[optind++];

    return 0;
}
